'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var varints = _interopDefault(require('varint'));
var withIs = _interopDefault(require('class-is'));

const toHex = d => d.reduce((hex, byte) => hex + byte.toString(16).padStart(2, '0'), '');
const fromHex = hex => {
  if (!hex.length) return new Uint8Array(0)
  return new Uint8Array(hex.match(/../g).map(b => parseInt(b, 16)))
};

const equals = (aa, bb) => {
  if (aa === bb) return true
  if (aa.byteLength !== bb.byteLength) {
    return false
  }

  for (let ii = 0; ii < aa.byteLength; ii++) {
    if (aa[ii] !== bb[ii]) {
      return false
    }
  }

  return true
};

const TypedArray = Object.getPrototypeOf(Int8Array);
const isTypedArray = obj => obj instanceof TypedArray;

const coerce = o => {
  if (o instanceof Uint8Array && o.constructor.name === 'Uint8Array') return o
  if (o instanceof ArrayBuffer) return new Uint8Array(o)
  if (o instanceof DataView || isTypedArray(o)) {
    return new Uint8Array(o.buffer, o.byteOffset, o.byteLength)
  }
  throw new Error('Unknown type, must be binary type')
};

const isBinary = o => {
  if (o instanceof DataView) return true
  if (o instanceof ArrayBuffer) return true
  if (isTypedArray(o)) return true
  return false
};

const fromString = str => (new TextEncoder()).encode(str);
const toString = b => (new TextDecoder()).decode(b);

var bytes = /*#__PURE__*/Object.freeze({
  __proto__: null,
  equals: equals,
  coerce: coerce,
  isBinary: isBinary,
  fromHex: fromHex,
  toHex: toHex,
  fromString: fromString,
  toString: toString
});

const readonly = (object, key, value) => {
  Object.defineProperty(object, key, {
    value,
    writable: false,
    enumerable: true
  });
};

var createCID = multiformats => {
  const { multibase, varint, multihash } = multiformats;
  const parse = buff => {
    const [code, length] = varint.decode(buff);
    return [code, buff.slice(length)]
  };
  const encode = (version, codec, multihash) => {
    return Uint8Array.from([
      ...varint.encode(version),
      ...varint.encode(codec),
      ...multihash
    ])
  };
  class CID {
    constructor (cid, ...args) {
      Object.defineProperty(this, '_baseCache', {
        value: new Map(),
        writable: false,
        enumerable: false
      });
      if (_CID.isCID(cid)) {
        readonly(this, 'version', cid.version);
        readonly(this, 'multihash', coerce(cid.multihash));
        readonly(this, 'buffer', coerce(cid.buffer));
        if (cid.code) readonly(this, 'code', cid.code);
        else readonly(this, 'code', multiformats.get(cid.codec).code);
        return
      }
      if (args.length > 0) {
        if (typeof args[0] !== 'number') throw new Error('String codecs are no longer supported')
        readonly(this, 'version', cid);
        readonly(this, 'code', args.shift());
        if (this.version === 0 && this.code !== 112) {
          throw new Error('Version 0 CID must be 112 codec (dag-cbor)')
        }
        this._multihash = args.shift();
        if (args.length) throw new Error('No longer supported, cannot specify base encoding in instantiation')
        if (this.version === 0) readonly(this, 'buffer', this.multihash);
        else readonly(this, 'buffer', encode(this.version, this.code, this.multihash));
        return
      }
      if (typeof cid === 'string') {
        if (cid.startsWith('Q')) {
          readonly(this, 'version', 0);
          readonly(this, 'code', 0x70);
          const { decode } = multibase.get('base58btc');
          this._multihash = decode(cid);
          readonly(this, 'buffer', this.multihash);
          return
        }
        const { name } = multibase.encoding(cid);
        this._baseCache.set(name, cid);
        cid = multibase.decode(cid);
      }
      cid = coerce(cid);
      readonly(this, 'buffer', cid);
      let code
      ;[code, cid] = parse(cid);
      if (code > 1) throw new Error(`Invalid CID version ${code}`)
      readonly(this, 'version', code)
      ;[code, cid] = parse(cid);
      readonly(this, 'code', code);
      this._multihash = cid;
    }

    set _multihash (hash) {
      const { length, digest } = multihash.decode(hash);
      if (digest.length !== length) throw new Error('Incorrect length')
      readonly(this, 'multihash', hash);
    }

    get codec () {
      throw new Error('"codec" property is deprecated, use integer "code" property instead')
    }

    get multibaseName () {
      throw new Error('"multibaseName" property is deprecated')
    }

    get prefix () {
      throw new Error('"prefix" property is deprecated')
    }

    toV0 () {
      if (this.code !== 0x70 /* dag-pb */) {
        throw new Error('Cannot convert a non dag-pb CID to CIDv0')
      }

      const { name } = multihash.decode(this.multihash);

      if (name !== 'sha2-256') {
        throw new Error('Cannot convert non sha2-256 multihash CID to CIDv0')
      }

      return new _CID(0, this.code, this.multihash)
    }

    toV1 () {
      return new _CID(1, this.code, this.multihash)
    }

    get toBaseEncodedString () {
      throw new Error('Deprecated, use .toString()')
    }

    [Symbol.for('nodejs.util.inspect.custom')] () {
      return 'CID(' + this.toString() + ')'
    }

    toString (base) {
      if (this.version === 0) {
        if (base && base !== 'base58btc') {
          throw new Error(`Cannot string encode V0 in ${base} encoding`)
        }
        const { encode } = multibase.get('base58btc');
        return encode(this.buffer)
      }
      if (!base) base = 'base32';
      if (this._baseCache.has(base)) return this._baseCache.get(base)
      this._baseCache.set(base, multibase.encode(this.buffer, base));
      return this._baseCache.get(base)
    }

    toJSON () {
      return {
        code: this.code,
        version: this.version,
        hash: this.multihash
      }
    }

    equals (other) {
      return this.code === other.code &&
        this.version === other.version &&
        equals(this.multihash, other.multihash)
    }
  }

  const _CID = withIs(CID, {
    className: 'CID',
    symbolName: '@ipld/js-cid/CID'
  });
  return _CID
};

const cache = new Map();
const varint = {
  decode: data => {
    const code = varints.decode(data);
    return [code, varints.decode.bytes]
  },
  encode: int => {
    if (cache.has(int)) return cache.get(int)
    const buff = Uint8Array.from(varints.encode(int));
    cache.set(int, buff);
    return buff
  }
};

const createMultihash = multiformats => {
  const { get, has, parse, add } = multiformats;
  const decode = digest => {
    const [info, len] = parse(digest);
    digest = digest.slice(len);
    const [length, len2] = varint.decode(digest);
    digest = digest.slice(len2);
    return { code: info.code, name: info.name, length, digest }
  };
  const encode = (digest, id) => {
    let info;
    if (typeof id === 'number') {
      info = { code: id };
    } else {
      info = get(id);
    }
    const code = varint.encode(info.code);
    const length = varint.encode(digest.length);
    return Uint8Array.from([...code, ...length, ...digest])
  };
  const hash = async (buff, key) => {
    buff = coerce(buff);
    const info = get(key);
    if (!info || !info.encode) throw new Error(`Missing hash implementation for "${key}"`)
    // https://github.com/bcoe/c8/issues/135
    /* c8 ignore next */
    return encode(await info.encode(buff), key)
  };
  const validate = async (_hash, buff) => {
    _hash = coerce(_hash);
    const { length, digest, code } = decode(_hash);
    if (digest.length !== length) throw new Error('Incorrect length')
    if (buff) {
      const { encode } = get(code);
      buff = await encode(buff);
      if (!equals(buff, digest)) throw new Error('Buffer does not match hash')
    }
    // https://github.com/bcoe/c8/issues/135
    /* c8 ignore next */
    return true
  };
  return { encode, has, decode, hash, validate, add, get }
};

const createMultibase = () => {
  const prefixMap = new Map();
  const nameMap = new Map();
  const _add = (prefix, name, encode, decode) => {
    prefixMap.set(prefix, [name, encode, decode]);
    nameMap.set(name, [prefix, encode, decode]);
  };
  const add = obj => {
    if (Array.isArray(obj)) {
      obj.forEach(add);
    } else {
      const { prefix, name, encode, decode } = obj;
      _add(prefix, name, encode, decode);
    }
  };
  const get = id => {
    if (id.length === 1) {
      if (!prefixMap.has(id)) throw new Error(`Missing multibase implementation for "${id}"`)
      const [name, encode, decode] = prefixMap.get(id);
      return { prefix: id, name, encode, decode }
    } else {
      if (!nameMap.has(id)) throw new Error(`Missing multibase implementation for "${id}"`)
      const [prefix, encode, decode] = nameMap.get(id);
      return { prefix, name: id, encode, decode }
    }
  };
  const has = id => {
    if (id.length === 1) {
      return prefixMap.has(id)
    }
    return nameMap.has(id)
  };
  const encode = (buffer, id) => {
    buffer = coerce(buffer);
    const { prefix, encode } = get(id);
    return prefix + encode(buffer)
  };
  const decode = string => {
    if (typeof string !== 'string') throw new Error('Can only multibase decode strings')
    const prefix = string[0];
    string = string.slice(1);
    if (string.length === 0) return new Uint8Array(0)
    const { decode } = get(prefix);
    return Uint8Array.from(decode(string))
  };
  const encoding = string => get(string[0]);
  return { add, has, get, encode, decode, encoding }
};

const create = (table = []) => {
  const intMap = new Map();
  const nameMap = new Map();
  const _add = (code, name, encode, decode) => {
    if (!Number.isInteger(code)) {
      throw new TypeError('multicodec entry must have an integer code')
    }
    if (typeof name !== 'string') {
      throw new TypeError('multicodec entry must have a string name')
    }
    if (encode != null && typeof encode !== 'function') {
      throw new TypeError('multicodec entry encode parameter must be a function')
    }
    if (decode != null && typeof decode !== 'function') {
      throw new TypeError('multicodec entry decode parameter must be a function')
    }
    intMap.set(code, [name, encode, decode]);
    nameMap.set(name, [code, encode, decode]);
  };
  for (const [code, name, encode, decode] of table) {
    _add(code, name, encode, decode);
  }
  const parse = buff => {
    buff = coerce(buff);
    const [code, len] = varint.decode(buff);
    let name, encode, decode;
    if (intMap.has(code)) {
[name, encode, decode] = intMap.get(code);
    }
    return [{ code, name, encode, decode }, len]
  };
  const get = obj => {
    if (typeof obj === 'string') {
      if (nameMap.has(obj)) {
        const [code, encode, decode] = nameMap.get(obj);
        return { code, name: obj, encode, decode }
      }
      throw new Error(`Do not have multiformat entry for "${obj}"`)
    }
    if (typeof obj === 'number') {
      if (intMap.has(obj)) {
        const [name, encode, decode] = intMap.get(obj);
        return { code: obj, name, encode, decode }
      }
      throw new Error(`Do not have multiformat entry for "${obj}"`)
    }
    if (isBinary(obj)) {
      return parse(coerce(obj))[0]
    }
    throw new Error('Unknown key type')
  };
  const has = id => {
    if (typeof id === 'string') {
      return nameMap.has(id)
    } else if (typeof id === 'number') {
      return intMap.has(id)
    }
    throw new Error('Unknown type')
  };
  // Ideally we can remove the coercion here once
  // all the codecs have been updated to use Uint8Array
  const encode = (value, id) => {
    const { encode } = get(id);
    return coerce(encode(value))
  };
  const decode = (value, id) => {
    const { decode } = get(id);
    return decode(coerce(value))
  };
  const add = obj => {
    if (Array.isArray(obj)) {
      obj.forEach(add);
    } else if (typeof obj === 'function') {
      add(obj(multiformats));
    } else {
      const { code, name, encode, decode } = obj;
      _add(code, name, encode, decode);
    }
  };

  const multiformats = { parse, add, get, has, encode, decode, varint, bytes };
  multiformats.multicodec = { add, get, has, encode, decode };
  multiformats.multibase = createMultibase();
  multiformats.multihash = createMultihash(multiformats);
  multiformats.CID = createCID(multiformats);
  return multiformats
};

exports.bytes = bytes;
exports.create = create;
exports.varint = varint;
